/*
	Copyright 2014 Gregorius van den Hoven - gregoriusvandenhoven@gmail.com
*/

/*
	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
//#include <sys/time.h>
#include <time.h>

#define BSC_X 256
#define BSC_Y 128
#define BSC_Z 64

#define BSC_L 2


// Kaze [
//#define compLT(x,y) (x < y)
//#define compGT(x,y) (x > y)

/* define the record to be sorted here */
#define LRECL 130 // Knight-Tours are 64x2+CRLF
//typedef int keyType;

typedef struct keyTypeTag {
        char data[LRECL];
} keyType;

#define compLT(x,y) (strcmpKAZE13(&x,&y)==-1)
#define compGT(x,y) (strcmpKAZE13(&x,&y)==1)

#define CRorLF 10 // 13 or 10
int strcmpKAZE13 (
        const char * src,
        const char * dst
        )
{
        int ret = 0 ;

        while( ! (ret = *(unsigned char *)src - *(unsigned char *)dst) && (*dst!=CRorLF))
                ++src, ++dst;

        if ( ret < 0 )
                ret = -1 ;
        else if ( ret > 0 )
                ret = 1 ;

        return( ret );
}
// Kaze ]


struct cube
{
	//int floor;
	keyType floor;
	short w_size;
	struct w_node **w_axis;
};

struct w_node
{
	//int floor;
	keyType floor;
	short x_size;
	struct x_node **x_axis;
};

struct x_node
{
	//int floor;
	keyType floor;
	short y_size;
	struct y_node **y_axis;
};

struct y_node
{
	//int floor;
	keyType floor;
	short z_size;
	struct z_node *z_axis;
};

struct z_node
{
	//int key;
	keyType key;
	int val;
};

void split_w_node(struct cube *cube, short w_index);
void split_x_node(struct cube *cube, short w_index, short x_index);
void split_y_node(struct cube *cube, short w_index, short x_index, short y_index);

//void insert_z_node(struct cube *cube, int key, int val);
void insert_z_node(struct cube *cube, keyType key, int val);

struct cube *create_cube(void)
{
	return (struct cube *) calloc(1, sizeof(struct cube));
}

void destroy_cube(struct cube *cube, struct z_node *z_array)
{
	struct w_node *w_node;
	struct x_node *x_node;
	struct y_node *y_node;

	register short w, x, y;

	register int z_size = 0;

	for (w = 0 ; w < cube->w_size ; w++)
	{
		w_node = cube->w_axis[w];

		for (x = 0 ; x < w_node->x_size ; x++)
		{
			x_node = w_node->x_axis[x];

			for (y = 0 ; y < x_node->y_size ; y++)
			{
				y_node = x_node->y_axis[y];

				memmove(&z_array[z_size], y_node->z_axis, y_node->z_size * sizeof(struct z_node));

				z_size += y_node->z_size;

				free(y_node->z_axis);
				free(y_node);
			}
			free(x_node->y_axis);
			free(x_node);
		}
		free(w_node->x_axis);
		free(w_node);
	}
	free(cube->w_axis);
	free(cube);
}

void insert_z_node(struct cube *cube, keyType key, int val)
{
	struct w_node *w_node;
	struct x_node *x_node;
	struct y_node *y_node;
	struct z_node *z_node;

	register short bot, top, w, x, y, z;

	if (cube->w_size == 0)
	{
		cube->w_axis = (struct w_node **) calloc(1, sizeof(struct w_node *));
		w_node = cube->w_axis[0] = (struct w_node *) calloc(1, sizeof(struct w_node));

		w_node->x_axis = (struct x_node **) calloc(1, sizeof(struct x_node *));
		x_node = w_node->x_axis[0] = (struct x_node *) calloc(1, sizeof(struct x_node));

		x_node->y_axis = (struct y_node **) calloc(1, sizeof(struct y_node *));
		y_node = x_node->y_axis[0] = (struct y_node *) calloc(1, sizeof(struct y_node));

		cube->w_size = w_node->x_size = x_node->y_size = 1;

		w = x = y = z = 0;

		cube->floor = w_node->floor = x_node->floor = y_node->floor = key;

		goto insert;
	}

	//if (key < cube->floor)
	if (compLT(key, cube->floor))
	{
		w_node = cube->w_axis[0];
		x_node = w_node->x_axis[0];
		y_node = x_node->y_axis[0];

		w = x = y = z = 0;

		cube->floor = w_node->floor = x_node->floor = y_node->floor = key;

		goto insert;
	}

	bot = 0;
	top = w = cube->w_size - 1;

	while (1)
	{
		//if (key < cube->w_axis[w]->floor)
		if (compLT(key, cube->w_axis[w]->floor))
		{
			top = w - 1;
		}
		else
		{
			bot = w;
		}

		if (bot + BSC_L > top)
		{
			break;
		}
		w = top - (top - bot) / 2;
	}

	for (w = top ; ; w--)
	{
		//if (key >= cube->w_axis[w]->floor)
		if (compLT(key, cube->w_axis[w]->floor))
		{} else
		{
			break;
		}
	}

	w_node = cube->w_axis[w];

	bot = 0;
	top = x = w_node->x_size - 1;

	while (1)
	{
		//if (key < w_node->x_axis[x]->floor)
		if (compLT(key, w_node->x_axis[x]->floor))
		{
			top = x - 1;
		}
		else
		{
			bot = x;
		}

		if (bot + BSC_L > top)
		{
			break;
		}
		x = top - (top - bot) / 2;
	}

	for (x = top ; ; x--)
	{
		//if (key >= w_node->x_axis[x]->floor)
		if (compLT(key, w_node->x_axis[x]->floor))
		{} else
		{
			break;
		}
	}

	x_node = w_node->x_axis[x];

	bot = 0;
	top = y = x_node->y_size - 1;

	while (1)
	{
		//if (key < x_node->y_axis[y]->floor)
		if (compLT(key, x_node->y_axis[y]->floor))
		{
			top = y - 1;
		}
		else
		{
			bot = y;
		}

		if (bot + BSC_L > top)
		{
			break;
		}
		y = top - (top - bot) / 2;
	}

	for (y = top ; ; y--)
	{
		//if (key >= x_node->y_axis[y]->floor)
		if (compLT(key, x_node->y_axis[y]->floor))
		{} else
		{
			break;
		}
	}

	y_node = x_node->y_axis[y];

	bot = 0;
	top = z = y_node->z_size - 1;

	while (1)
	{
		//if (key < y_node->z_axis[z].key)
		if (compLT(key, y_node->z_axis[z].key))
		{
			top = z - 1;
		}
		else
		{
			bot = z;
		}

		if (bot + BSC_L > top)
		{
			break;
		}

		z = top - (top - bot) / 2;
	}

	for (z = top ; ; z--)
	{
		//if (key >= y_node->z_axis[z].key)
		if (compLT(key, y_node->z_axis[z].key))
		{} else
		{
			break;
		}
	}

	z++;

	insert:

	y_node->z_size++;

	y_node->z_axis = (struct z_node *) realloc(y_node->z_axis, y_node->z_size * sizeof(struct z_node));

	if (z + 1 != y_node->z_size)
	{
		memmove(&y_node->z_axis[z + 1], &y_node->z_axis[z], (y_node->z_size - z - 1) * sizeof(struct z_node));
	}

	z_node = &y_node->z_axis[z];

	z_node->key = key;
	z_node->val = val;

	if (y_node->z_size >= BSC_Z)
	{
		split_y_node(cube, w, x, y);

		if (x_node->y_size > BSC_Y)
		{
			split_x_node(cube, w, x);

			if (w_node->x_size > BSC_X)
			{
				split_w_node(cube, w);
			}
		}
	}
}


void insert_w_node(struct cube *cube, short w_index)
{
	cube->w_size++;

	cube->w_axis = (struct w_node **) realloc(cube->w_axis, cube->w_size * sizeof(struct w_node *));

	if (w_index + 1 != cube->w_size)
	{
		memmove(&cube->w_axis[w_index + 1], &cube->w_axis[w_index], (cube->w_size - w_index - 1) * sizeof(struct w_node *));
	}

	cube->w_axis[w_index] = (struct w_node *) calloc(1, sizeof(struct w_node));
}

void split_w_node(struct cube *cube, short w_index)
{
	struct w_node *w_node1, *w_node2;

	insert_w_node(cube, w_index + 1);

	w_node1 = cube->w_axis[w_index];
	w_node2 = cube->w_axis[w_index + 1];

	w_node2->x_size = w_node1->x_size / 2;
	w_node1->x_size = w_node1->x_size - w_node2->x_size;

	w_node2->x_axis = (struct x_node **) realloc(w_node2->x_axis, w_node2->x_size * sizeof(struct x_node *));

	memcpy(&w_node2->x_axis[0], &w_node1->x_axis[w_node1->x_size], w_node2->x_size * sizeof(struct x_node *));

	w_node1->x_axis = (struct x_node **) realloc(w_node1->x_axis, w_node1->x_size * sizeof(struct x_node *));

	w_node2->floor = w_node2->x_axis[0]->floor;
}


void insert_x_node(struct cube *cube, struct w_node *w_node, short x_index)
{
	w_node->x_size++;

	w_node->x_axis = (struct x_node **) realloc(w_node->x_axis, w_node->x_size * sizeof(struct x_node *));

	if (w_node->x_size != x_index + 1)
	{
		memmove(&w_node->x_axis[x_index + 1], &w_node->x_axis[x_index], (w_node->x_size - x_index - 1) * sizeof(struct x_node *));
	}

	w_node->x_axis[x_index] = (struct x_node *) calloc(1, sizeof(struct x_node));
}

void split_x_node(struct cube *cube, short w_index, short x_index)
{
	struct w_node *w_node = cube->w_axis[w_index];
	struct x_node *x_node1, *x_node2;

	insert_x_node(cube, w_node, x_index + 1);

	x_node1 = w_node->x_axis[x_index];
	x_node2 = w_node->x_axis[x_index + 1];

	x_node2->y_size = x_node1->y_size / 2;
	x_node1->y_size = x_node1->y_size - x_node2->y_size;

	x_node2->y_axis = (struct y_node **) realloc(x_node2->y_axis, x_node2->y_size * sizeof(struct y_node *));

	memcpy(&x_node2->y_axis[0], &x_node1->y_axis[x_node1->y_size], x_node2->y_size * sizeof(struct y_node *));

	x_node1->y_axis = (struct y_node **) realloc(x_node1->y_axis, x_node1->y_size * sizeof(struct y_node *));

	x_node2->floor = x_node2->y_axis[0]->floor;
}


void insert_y_node(struct cube *cube, struct x_node *x_node, short y_index)
{
	x_node->y_size++;

	x_node->y_axis = (struct y_node **) realloc(x_node->y_axis, x_node->y_size * sizeof(struct y_node *));

	if (x_node->y_size != y_index + 1)
	{
		memmove(&x_node->y_axis[y_index + 1], &x_node->y_axis[y_index], (x_node->y_size - y_index - 1) * sizeof(struct y_node *));
	}

	x_node->y_axis[y_index] = (struct y_node *) calloc(1, sizeof(struct y_node));
}

void split_y_node(struct cube *cube, short w_index, short x_index, short y_index)
{
	struct x_node *x_node = cube->w_axis[w_index]->x_axis[x_index];
	struct y_node *y_node1, *y_node2;

	insert_y_node(cube, x_node, y_index + 1);

	y_node1 = x_node->y_axis[y_index];
	y_node2 = x_node->y_axis[y_index + 1];

	y_node2->z_size = y_node1->z_size / 2;
	y_node1->z_size = y_node1->z_size - y_node2->z_size;

	y_node2->z_axis = (struct z_node *) realloc(y_node2->z_axis, y_node2->z_size * sizeof(struct z_node));

	memcpy(&y_node2->z_axis[0], &y_node1->z_axis[y_node1->z_size], y_node2->z_size * sizeof(struct z_node));

	y_node1->z_axis = (struct z_node *) realloc(y_node1->z_axis, y_node1->z_size * sizeof(struct z_node));

	y_node2->floor = y_node2->z_axis[0].key;
}


/*
long long utime()
{
	struct timeval now_time;

	gettimeofday(&now_time, NULL);

	return now_time.tv_sec * 1000000LL + now_time.tv_usec;
}
*/

int main(int argc, char **argv)
{
	int cnt, max;
	long long start, end;
	struct z_node *z_array;
	struct cube *cube;

FILE *ifp;

        if ((ifp = fopen(argv[1], "rb")) == NULL) {
            printf("Error: file %s, unable to open.\n", argv[1]);
            return(13);
        }
/*
	if (argv[1] && *argv[1])
	{
		printf("%s\n", argv[1]);
	}
*/

	cube = create_cube();

	//max = 5000000;
	max = 10000000;

	z_array = malloc(max * sizeof(struct z_node));

/*
	for (cnt = 0 ; cnt < max / 2 ; cnt++)
	{
		z_array[cnt].key = rand();
		z_array[cnt].val = cnt;
	}

	for (cnt = max / 2 ; cnt < max ; cnt++)
	{
		z_array[cnt].key = cnt + 1000000000;
		z_array[cnt].val = cnt;
	}
*/

	for (cnt = 0 ; cnt < max ; cnt++)
	{
		//z_array[cnt].key = rand();
		fread(&z_array[cnt].key, sizeof(keyType), 1, ifp);
		z_array[cnt].val = cnt;
	}

	//start = utime();
	start = clock();

	for (cnt = 0 ; cnt < max ; cnt++)
	{
		insert_z_node(cube, z_array[cnt].key, z_array[cnt].val);
	}

	destroy_cube(cube, z_array);

	//end = utime();
	end = clock();


	for (cnt = 1 ; cnt < max ; cnt++)
	{
		//if (z_array[cnt - 1].key > z_array[cnt].key)
		if (compGT(z_array[cnt - 1].key, z_array[cnt].key))
		{
			printf("Cubesort: not properly sorted at index %d.\n", cnt);
			return 0;
		}
	}

	//printf("Cubesort: sorted %d elements in %f seconds.\n", max, (end - start) / 1000000.0);
	printf("Cubesort: sorted %d elements in %d clocks.\n", max, (end - start));

// Dumping:
	fclose(ifp);
        if ((ifp = fopen("Cubesort_ResultantFile.txt", "wb")) == NULL) {
            printf("Error: file %s, unable to open.\n", "Cubesort_ResultantFile.txt");
            return(1113);
        }
	for (cnt = 0 ; cnt < max ; cnt++)
	{
		fwrite(&z_array[cnt].key, sizeof(keyType), 1, ifp);
	}
	fclose(ifp);

	return 0;
}
